
# Juju requires higher frame size for large models
MAX_FRAME_SIZE = 2**26

# Not all charms use the openstack-origin. The openstack specific
# charms do, but some of the others use an alternate origin key
# depending on who the author was.
ORIGIN_KEYS = {
    'ceph': 'source',
    'ceph-osd': 'source',
    'ceph-mon': 'source',
    'ceph-radosgw': 'source',
}

# Default list of services, used in upgrade if apps not specified in params
# Services are upgraded in the order specified
# The order of services is based on:
# https://github.com/openstack-charmers/openstack-charms-tools/blob/master/os-upgrade.py
# https://docs.openstack.org/project-deploy-guide/charm-deployment-guide/latest/app-upgrade-openstack.html
SERVICES = [
    # Identity
    'keystone',

    # Ceph
    'ceph-mon',
    'ceph-osd',
    'ceph-radosgw',

    # Image
    'glance',

    # Upgrade nova
    'nova-cloud-controller',

    'nova-compute',

    # Neutron upgrades
    'neutron-api',

    'neutron-gateway',

    'neutron-openvswitch',

    # Backend block-storage upgrade.
    # Note: just upgrade cinder service.
    'cinder',
    'cinder-ceph',

    # Upgrade dashboard
    'openstack-dashboard',

    'rabbitmq-server',
]



#!/usr/bin/python3

import asyncio
import websockets
import logging
from collections import Counter
from jujuna.helper import connect_juju, log_traceback
from juju.errors import JujuError


# create logger
log = logging.getLogger('jujuna.deploy')


async def deploy(
    bundle_file,
    ctrl_name='',
    model_name='',
    wait=False,
    endpoint='',
    username='',
    password='',
    cacert='',
    **kwargs
):
    """Deploy a local juju bundle.

    Handles deployment of a bundle file to the current or selected model.

    Connection requires juju client configs to be present locally or specification of credentialls:
    endpoint (e.g. 127.0.0.1:17070), username, password, and model uuid as model_name.

    :param bundle_file: juju bundle file
    :param ctrl_name: juju controller
    :param model_name: juju model name or uuid
    :param wait: boolean
    :param endpoint: string
    :param username: string
    :param password: string
    :param cacert: string
    """
    log.info('Reading bundle: {}'.format(bundle_file.name))
    entity_url = 'local:' + bundle_file.name.replace('/bundle.yaml', '')

    controller, model = await connect_juju(
        ctrl_name,
        model_name,
        endpoint=endpoint,
        username=username,
        password=password,
        cacert=cacert
    )

    try:
        # Deploy a bundle
        log.info("Deploy: {}".format(entity_url))
        deployed_apps = await model.deploy(
            entity_url
        )

        if wait:
            await wait_until(
                model,
                deployed_apps,
                loop=model.loop
            )
        try:
            d = Counter([a.status for a in deployed_apps])
        except Exception:
            log.error('Collecting status failed')
            d = {}
        log.info('{} - Machines: {} Apps: {} Stats: {}'.format(
            'DEPLOYED' if wait else 'CURRENT',
            len(model.machines),
            len(model.applications),
            dict(d)
        ))

    except JujuError as e:
        log.error('JujuError during deploy')
        log_traceback(e)
    finally:
        # Disconnect from the api server and cleanup.
        await model.disconnect()
        await controller.disconnect()


async def wait_until(model, apps, log_time=5, timeout=None, wait_period=0.5, loop=None):
    """Blocking with logs.

    Return only after all conditions are true.

    :param model: juju model
    :param apps: juju application
    :param log_time: logging frequency (s)
    :param timeout: blocking timeout (s)
    :param wait_period: waiting time between checks (s)
    :param loop: asyncio event loop
    """
    log_count = 0

    def _disconnected():
        return not (model.is_connected() and model.connection().is_open)

    async def _block(log_count):
        blockable = ['maintenance', 'blocked', 'waiting', 'error']
        while (
            not _disconnected() and
            any(a.status != 'active' for a in apps) and
            any(u.workload_status in blockable for a in apps for u in a.units)
        ):
            await asyncio.sleep(wait_period, loop=loop)
            log_count += 0.5
            if log_count % log_time == 0:
                ass = Counter([
                    a.status for a in apps
                ])
                wss = Counter([
                    unit.workload_status for a in apps for unit in a.units
                ])
                log.info('PROGRESS - Machines: {} Apps: {} Stats: {} Workload: {}'.format(
                    len(model.machines),
                    len(model.applications),
                    dict(ass),
                    dict(wss)
                ))
    await asyncio.wait_for(_block(log_count), timeout, loop=loop)

    if _disconnected():
        raise websockets.ConnectionClosed(1006, 'no reason')



"""Process broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter


class Process(Broker):
    """Process broker."""

    def __init__(self):
        """Init Process broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            act = await unit.run(python3(exporter), timeout=10)
            results = load_output(act.data['results'])
            # print(results)
            for condition in test_case:
                rows.append((idx, '{} == present'.format(condition), condition in results), )

        return rows



"""Service broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter


class Service(Broker):
    """Service broker."""

    def __init__(self):
        """Init service broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            act = await unit.run(python3(exporter), timeout=10)
            results = load_output(act.data['results'])
            # print(results['services'])
            # print(test_case)
            for service, condition in test_case.items():
                rows.append((idx, '{} == {}'.format(service, 'exists'), service in results['services']), )
                for c, v in condition.items():
                    rows.append((
                        idx, '{}.{} == {}'.format(service, c, v),
                        service in results['services'] and results['services'][service][c] == v
                    ), )

        return rows



import time  # noqa
import yaml
import json
import logging
import async_timeout
from collections import defaultdict
from juju.errors import JujuError
from jujuna.helper import connect_juju, log_traceback

from jujuna.brokers.api import Api as ApiBroker
from jujuna.brokers.file import File as FileBroker
from jujuna.brokers.mount import Mount as MountBroker
from jujuna.brokers.network import Network as NetworkBroker
from jujuna.brokers.package import Package as PackageBroker
from jujuna.brokers.process import Process as ProcessBroker
from jujuna.brokers.service import Service as ServiceBroker
from jujuna.brokers.user import User as UserBroker


# create logger
log = logging.getLogger('jujuna.tests')


async def test(
    test_suite='',
    ctrl_name='',
    model_name='',
    endpoint='',
    username='',
    password='',
    cacert='',
    **kwargs
):
    """Run a test suite against applications deployed in the current or selected model.

    Applications are tested with declarative parameters specified in the test suite using the available brokers.

    Connection requires juju client configs to be present locally or specification of credentialls:
    endpoint (e.g. 127.0.0.1:17070), username, password, and model uuid as model_name.

    :param test_suite: suite file (Yaml)
    :param ctrl_name: juju controller
    :param model_name: juju model name or uuid
    :param endpoint: string
    :param username: string
    :param password: string
    :param cacert: string
    """
    if test_suite:
        with open(test_suite.name, 'r') as stream:
            try:
                suite = yaml.load(stream)
            except yaml.YAMLError as exc:
                log.error(exc)

    controller, model = await connect_juju(
        ctrl_name,
        model_name,
        endpoint=endpoint,
        username=username,
        password=password,
        cacert=cacert
    )

    model_passed, model_failed = 0, 0
    failed_units = set()

    try:
        for app_name, app in model.applications.items():
            if suite and app_name in suite:
                app_passed, app_failed = 0, 0
                log.info('[{}]: {} {} [{}]'.format(app_name, app.status, app.alive, len(app.units)))
                for idx, unit in enumerate(app.units):
                    async with async_timeout.timeout(60):
                        passed, failed = await execute_brokers(suite[app_name], unit, idx)
                        app_passed += passed
                        app_failed += failed
                        if app.status in ['error', 'maintenance', 'blocked'] or failed:
                            failed_units.add(unit.name)
                model_passed += app_passed
                model_failed += app_failed
                log.info('[{}]: Passed: {} Failed: {}'.format(app_name, app_passed, app_failed))

        alive = defaultdict(int)
        status = defaultdict(int)
        for app in model.applications.values():
            alive[app.alive] += 1
            status[app.status] += 1

        if (
            False not in alive and
            all(s in ['active', 'waiting'] for s in status.keys())
        ):
            log.info('All juju apps state to be alive and active.')
        else:
            log.info('Finished with errors: Alive: {} Status: {}'.format(dict(alive), dict(status)))

        log.info('[FINISHED] Passed tests: {} Failed tests: {}'.format(model_passed, model_failed))

    except JujuError as e:
        log.error('JujuError during tests')
        log_traceback(e)
    finally:
        # Disconnect from the api server and cleanup.
        await model.disconnect()
        await controller.disconnect()

    if failed_units:
        log.error('Failed units: {}'.format(', '.join(sorted(failed_units))))
        return 1


def python3(file):
    return 'python3 {}'.format(file)


def load_output(results):
    """Load json result from broker.

    """
    if results['Code'] == '0':
        try:
            var = json.loads(results['Stdout'])
        except Exception as e:
            raise Exception('JSON load failed: {}'.format(e))
    else:
        raise Exception('Operation failed: {}'.format(results['Stderr']))
    return var


async def execute_brokers(app_test_suite, unit, idx):
    """Iterate brokers and acquire results.

    """
    broker_map = {
        'api': ApiBroker,
        'file': FileBroker,
        'mount': MountBroker,
        'network': NetworkBroker,
        'package': PackageBroker,
        'process': ProcessBroker,
        'service': ServiceBroker,
        'user': UserBroker,
    }
    passed = 0
    failed = 0

    for test_case in app_test_suite.keys():
        if test_case in broker_map.keys():
            rows = await broker_map[test_case]().run(app_test_suite[test_case], unit, idx)
            for row in rows:
                log.info('[{}]: {} {} [{}]'.format(unit.entity_id, test_case, row[1], "Pass" if row[2] else "Fail"))
                if row[2]:
                    passed += 1
                else:
                    failed += 1
        else:
            log.info("TEST: Skipped (Broker '{}' not registered)".format(test_case))
    return passed, failed



"""Exporter management."""

import os
import uuid
import async_timeout


class Exporter():
    """Exporter context manager."""

    def __init__(self, unit, exporter_name):
        """Init exporter."""
        if exporter_name.endswith('.py'):
            raise Exception('Do not provide an extension')
        self.unit = unit
        self.full_path = os.path.dirname(os.path.realpath(__file__))
        self.exporter_local = os.path.join(self.full_path, exporter_name + '.py')
        if not os.path.isfile(self.exporter_local):
            raise Exception('Exporter: {}.py not found'.format(exporter_name))
        self.exporter_remote = os.path.join('/tmp', str(uuid.uuid4())[:9] + exporter_name + '.py')

    async def __aenter__(self):
        """Upload exporter and return remote path."""
        # print('Load: ', self.exporter_local)
        await self._load_exporter(self.unit, self.exporter_local, self.exporter_remote)
        return self.exporter_remote

    async def __aexit__(self, type, value, traceback):
        """Dispose remote exporter."""
        # print('Unload: ', self.exporter_remote)
        await self._unload_exporter(self.unit, self.exporter_remote)

    async def _load_exporter(self, unit, source, target, user='ubuntu'):
        """Upload exporter."""
        async with async_timeout.timeout(15):
            await unit.scp_to(
                source,
                target,
                user=user
            )

    async def _unload_exporter(self, unit, target, user='ubuntu'):
        """Dispose of remote exporter."""
        for i in range(3):
            try:
                async with async_timeout.timeout(15):
                    ret = await unit.run('rm {}'.format(target), timeout=10)
                break
            except Exception as e:
                # Try 3 times, if it fails raise on last
                if i == 2 and 'No such file' not in str(e):
                    raise e
        # ret = await unit.run('rm /tmp/*.py', timeout=3)
        if (
            ret and ret.data and
            'results' in ret.data and
            ret.data['results']['Code'] != '0' and
            'No such file' not in ret.data['results']['Stderr']
        ):
            raise Exception('Unload failed: ', ret.data['results']['Stderr'])



#!/usr/bin/python3

import asyncio
import websockets
import logging
from jujuna.helper import connect_juju
from juju.errors import JujuError


# create logger
log = logging.getLogger('jujuna.clean')


async def wait_until(model, *conditions, log_time=5, timeout=None, wait_period=0.5, loop=None):
    """Return only after all conditions are true.

    """
    log_count = 0

    def _disconnected():
        return not (model.is_connected() and model.connection().is_open)

    async def _block(log_count):
        while not _disconnected() and not all(c() for c in conditions):
            await asyncio.sleep(wait_period, loop=loop)
            log_count += 0.5
            if log_count % log_time == 0:
                log.info('[RUNNING] Machines: {} {} Apps: {}'.format(
                    len(model.machines),
                    ', '.join(model.machines.keys()),
                    len(model.applications)
                ))
    await asyncio.wait_for(_block(log_count), timeout, loop=loop)

    if _disconnected():
        raise websockets.ConnectionClosed(1006, 'no reason')

    log.info('[DONE] Machines: {} Apps: {}'.format(
        len(model.machines),
        len(model.applications)
    ))


async def clean(
    ctrl_name='',
    model_name='',
    ignore=[],
    wait=False,
    force=False,
    dry_run=False,
    endpoint='',
    username='',
    password='',
    cacert='',
    **kwargs
):
    """Destroy applications present in the current or selected model.

    Connection requires juju client configs to be present locally or specification of credentialls:
    endpoint (e.g. 127.0.0.1:17070), username, password, and model uuid as model_name.

    :param ctrl_name: juju controller
    :param model_name: juju model name or uuid
    :param ignore: list of application names
    :param wait: boolean
    :param force: boolean
    :param dry_run: boolean
    :param endpoint: string
    :param username: string
    :param password: string
    :param cacert: string
    """
    controller, model = await connect_juju(
        ctrl_name,
        model_name,
        endpoint=endpoint,
        username=username,
        password=password,
        cacert=cacert
    )

    try:
        # Remove all the apps
        for app in model.applications:
            if app in ignore:
                log.info('Ignoring {}'.format(app))
            else:
                log.info('Remove {} from model'.format(app))
                if not dry_run:
                    await model.applications[app].destroy()

        if not ignore and force:
            machines = [m for m in model.machines.values() if 'arch' in m.safe_data['hardware-characteristics']]
            for machine in machines:
                log.info('Remove machine {} from model'.format(machine.entity_id))
                if not dry_run:
                    try:
                        await machine.destroy(force=True)
                    except Exception as e:
                        log.warn('ERROR: {}'.format(e))

        if wait and not ignore and not dry_run:
            await wait_until(
                model,
                lambda: model.applications == {} and model.machines == {},
                loop=model.loop
            )

    except JujuError as e:
        log.info(e.message)
    finally:
        # Disconnect from the api server and cleanup.
        await model.disconnect()
        await controller.disconnect()



"""User broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter


class User(Broker):
    """User broker."""

    def __init__(self):
        """Init User broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            act = await unit.run(python3(exporter), timeout=10)
            user_data = load_output(act.data['results'])
            # print(user_data)
            for condition, test_item in test_case.items():
                test_res = False
                # if subitem in test_item.items():
                if all(user_data[condition][key] == value for key, value in test_item.items()):
                    test_res = True
                rows.append((idx, '{} == {}'.format(condition, 'present'), test_res), )

        return rows



#!/usr/bin/env python3

import re
import sys
import json
import base64


def decode_address(address):
    hex_ip, hex_port = address.split(':')
    ip = '.'.join([str(item) for item in reversed(list(base64.b16decode(hex_ip)))])
    port = int.from_bytes(base64.b16decode(hex_port), byteorder='big')
    return ip, port


def sockets():
    with open('/proc/net/tcp', 'r') as f:
        lines = []
        for line in f:
            line = line.strip()
            line = re.sub(' +', ' ', line)
            lines.append(line.split())

    sock_list = []
    header = lines[0]

    for line in lines[1:]:
        line_struct = {}
        for idx, head in enumerate(header):
            line_struct[head] = line[idx]
        sock_list.append(line_struct)

    data_list = []
    for item in sock_list:
        data_item = {}
        data_item['local_ip'], data_item['local_port'] = decode_address(item['local_address'])
        data_item['rem_ip'], data_item['rem_port'] = decode_address(item['rem_address'])
        data_list.append(data_item)

    return data_list


def main():
    data = {
        'sockets': sockets(),
    }

    print(json.dumps(data))
    sys.exit(0)


if __name__ == "__main__":
    main()



import logging
import traceback

from jujuna.settings import MAX_FRAME_SIZE

from juju.controller import Controller
from juju.model import Model

from theblues.charmstore import CharmStore


def log_traceback(ex, prefix=''):
    if prefix and isinstance(prefix, str):
        prefix = '{} - '.format(prefix.strip())
    ex_traceback = ex.__traceback__
    for line in traceback.format_exception(
        ex.__class__, ex, ex_traceback
    ):
        logging.error('{}{}'.format(prefix, line.rstrip('\n')))


def cs_name_parse(name, series=None):
    """Charm store name parse.

    """
    try:
        cscode, csuri = name.split(':', 1)
        items = csuri.split('/') if '/' in csuri else ['', csuri]
        return {
            'series': items[0] if items[0] or not series else series,
            'charm': '-'.join(items[1].split('-')[:-1]),
            'revision': int(items[1].split('-')[-1]),
            'charmstore': False if cscode == 'local' else True
        }
    except Exception:
        logging.warn('Failed to parse Charm name: {}'.format(name))
        return {}


async def connect_juju(ctrl_name=None, model_name=None, endpoint=None, username=None, password=None, cacert=None):
    controller = Controller(max_frame_size=MAX_FRAME_SIZE)  # noqa

    if endpoint:
        await controller.connect(endpoint=endpoint, username=username, password=password, cacert=cacert)
    else:
        await controller.connect(ctrl_name)

    if endpoint:
        model = Model(max_frame_size=MAX_FRAME_SIZE)
        await model.connect(uuid=model_name, endpoint=endpoint, username=username, password=password, cacert=cacert)
    elif model_name:
        model = await controller.get_model(model_name)
    else:
        model = Model(max_frame_size=MAX_FRAME_SIZE)  # noqa
        await model.connect()

    # HACK low unsettable timeout in the model
    model.charmstore._cs = CharmStore(timeout=60)

    return controller, model



#!/usr/bin/env python3

import json
import sys
import apt


def main():
    pkg_data = {}

    cache = apt.Cache()
    installed = {}
    for mypkg in apt.Cache():
        if cache[mypkg.name].is_installed:
            installed[mypkg.name] = {
                'id': mypkg.id,
                'name': mypkg.name,
                'shortname': mypkg.shortname,
                'versions': list(dict(mypkg.versions).keys())
            }

    pkg_data = {
        'installed': installed
    }

    print(json.dumps(pkg_data))
    sys.exit(0)


if __name__ == "__main__":
    main()



"""Network broker."""
from . import Broker, python3, load_output
from jujuna.exporters import Exporter


class Network(Broker):
    """Network broker."""

    def __init__(self):
        """Init broker."""
        super().__init__()

    async def run(self, test_case, unit, idx):
        """Run tests."""
        rows = []
        async with Exporter(unit, self.named) as exporter:
            if 'port' in test_case:
                test_data = await unit.run(python3(exporter), timeout=10)
                results = load_output(test_data.data['results'])
                # print(results)

                local_ports = [s['local_port'] for s in results['sockets']]

                for port in test_case['port']:
                    rows.append((idx, '{}.{} == {}'.format('port', port, 'open'), port in local_ports), )

        return rows
